#! /usr/bin/env python
# -*- coding: iso-8859-15 -*-
#
# Simple CSV parser.
# Copyright (C) 2013 MiKael NAVARRO
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""Simple CSV parser.
Copyright (C) 2013 MiKael NAVARRO

Example on how interprets CSV file into dictionary (first column is the key).

"""

# Specific variables for pydoc
__author__ = "MiKael Navarro <klnavarro@gmail.com>"
__date__ = "Thu Dec 3 2013"
__version__ = "0.1.0"
__credits__ = """https://pypi.python.org/pypi/pyparsing/1.5.7,
https://pypi.python.org/pypi/ordereddict"""


# Include directives
import sys

from decimal import Decimal, InvalidOperation  # numbers represented exactly

from pprint import pprint as pp

HAS_PYPARSING = True
try:
    from pyparsing import *
except ImportError:
    HAS_PYPARSING = False

HAS_ORDEREDDICT = True
try:
    from collections import OrderedDict
except ImportError:
    try:
        from ordereddict import OrderedDict
    except ImportError:
        HAS_ORDEREDDICT = False


# Global variables
BNF="""
line
    key;val1;val2;val3;...
val
    "string"
    number
"""


#
# Convert value to right type.
#
def conv_value(val):
    """Convert value to int|float|str."""

    try:
        return int(val)
    except ValueError, ve:
        try:
            return Decimal(val)
        except InvalidOperation, ve:
            return str(val)


#
# Convert token (pyparsing.ParserElement.setParseAction callback).
#
def conv_number(s, l, toks):
    """Convert token to int|float|str (pyparsing.ParserElement.setParseAction callback)."""

    return conv_value(toks[0])


#
# Parse DATA into dictionary.
#
def csv_parse(data, ordered=False):
    """Parse data into dictionary.

    >>> sample_data = '''
    ... #key;val1;val2
    ... a;int;1
    ... b;int;2
    ... c;int;3
    ... tab[0];int;0
    ... tab[1];int;1
    ... tab[2];int;2
    ... tab[3];int;3
    ... const.pi.int;int;3
    ... const.pi.float;float;3.14
    ... const.pi.str;str;"pi"
    ... const.e;float;2.7182818284590451
    ... d;int;4
    ... '''
    >>> results = csv_parse(sample_data)
    >>> pp(results)
    {'a': ['int', 1],
     'b': ['int', 2],
     'c': ['int', 3],
     'const.e': ['float', Decimal("2.7182818284590451")],
     'const.pi.float': ['float', Decimal("3.14")],
     'const.pi.int': ['int', 3],
     'const.pi.str': ['str', '"pi"'],
     'd': ['int', 4],
     'tab[0]': ['int', 0],
     'tab[1]': ['int', 1],
     'tab[2]': ['int', 2],
     'tab[3]': ['int', 3]}
    """

    if HAS_ORDEREDDICT and ordered:
        fields = OrderedDict()
    else:
        fields = {}  # unordered

    if HAS_PYPARSING:  # parse data into pyparsing.ParseResults object

        # Define line-related parts
        EOL = LineEnd().suppress()
        SOL = LineStart().leaveWhitespace()

        # Comment = # + ...
        comment = ('#' + restOfLine).suppress()
        # Empty line
        blankline = SOL + LineEnd()

        # Use parse actions to do parse-time conversion of values
        real = Combine(Optional('-') + ('0' | Word('123456789', nums)) + '.' + Word(nums))
        #real = Regex(r"[+-]?\d+\.\d*").setParseAction(lambda t: float(t[0]))
        integer = Word(nums)
        #integer = Regex(r"[+-]?\d+").setParseAction(lambda t: int(t[0]))
        number = real | integer
        #number = Combine(Optional('-') + ('0' | Word('123456789', nums)) + Optional('.' + Word(nums)))
        number.setParseAction(conv_number)

        # Parses a string enclosed in quotes, but strips off the quotes at parse time
        #string = QuotedString('"')
        #string = Regex(r"[a-zA-Z0-9_.\[\]'\"]+")
        string = Word(alphas+"_'\"", alphanums+"_.[]'\"")

        # Line = semi-column separated values
        #line = commaSeparatedList
        bnf = delimitedList(string | number, delim=';')

        # Ignore blank lines and comments
        bnf.ignore(blankline)
        bnf.ignore(comment)

        # Scan data... into dict
        for tokens, start, end in bnf.scanString(data):
            fields[tokens[0]] = tokens[1:]
            
    else:
        
        # Split lines... into list
        #lines = []
        #for l in data.splitlines():
        #    if l and not l.startswith('#'):
        #        lines.append(l.split(';'))
        #
        #if __debug__:
        #    pp(lines)

        # Split lines... into dict
        for l in data.splitlines():
            if l and not l.startswith('#'):
                line = l.split(';')
                fields[line[0]] = line[1:]

    if __debug__:
        pp(fields)
    return fields


#
# Main entry point.
#
def main(options, csvfile):
    """Parse CSV file."""

    # Get data from file
    data = open(csvfile, 'r').read()

    # Parse data
    results = csv_parse(data, options.ordered)
    for k, v in results.items():
        print k, v


#
# External entry point.
#
if __name__ == "__main__":
    # Get options
    from optparse import OptionParser, make_option

    option_list = [
        make_option("-o", "--ordered",
                    action="store_true", dest="ordered", 
                    default=False,
                    help="keep order if possible"),
        make_option("-t", "--test",
                    action="store_true", dest="test", 
                    default=False,
                    help="sample test"),
        ]

    parser = OptionParser(usage="python -O %prog [options] ... [args] ...", 
                          version=__version__,
                          description="Interprets CSV file into dictionary.",
                          option_list=option_list)

    (options, args) = parser.parse_args()

    if options.test:  # self-tests
        import doctest
        doctest.testmod(sys.modules[__name__])
        #doctest.run_docstring_examples(parse, globals())
    else:
        # Needed argument
        if not args:
            parser.error("missing file argument")
        elif len(args) != 1:
            parser.error("incorrect number of arguments")
        
        # Process start here
        main(options, args[0])


